/* xmitrc.c Radio Control transmitter code for T-Zero Battle bot
Ver. 1.0

By Dale Heatherington, WA4DSY
dale.h@wa4dsy.net

Atmel AVR AT90S8535 microcontroller

Compiled with avr-gcc  (See Makefile)


-------------------------
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA
------------------------

*/
 

/***************************************************************************/

//Note: CPU type defined in Makefile

#include <io.h>
#include <sig-avr.h>
#include <interrupt.h>
#include <pgmspace.h>
#include <progmem.h>


                
#define F_CPU    7372800        /* crystal frequency of CPU */

#define COUNT 1      /* offset to count field in tx/rx buffers */
#define ADDRESS 0    /* offset to address field in tx/rx buffers */
#define DATA 2       /* offset to start of data in tx/rx buffers */
#define LED_RED 2
#define LED_YELLOW 3

/* Switch Matrix bit defs */

/*matrix0*/

#define m_mode 0x80
#define m_east 0x40
#define m_north 0x20
#define m_rumble 0x10
#define m_south 0x08
#define m_s 0x04
#define m_left_fire 0x02
#define m_west  0x01

/*matrix1*/

#define m_right_fire 0x80
#define m_A          0x40
#define m_B          0x20
#define m_C          0x10
#define m_Z          0x08
#define m_Y          0x04
#define m_X          0x20





uint8_t matrix0,matrix1;
uint16_t ad0,ad1,ad2,ad3,ad4;   //A-D converter values



uint8_t tx_buf[16];      //Serial tx buffer
uint8_t tx_ptr,tx_ctr ;
uint16_t crc;
int8_t txstate, counter,val;

int ii;
unsigned char prescale ;
unsigned char prescale100;
unsigned char prescale1sec;
unsigned char prescaleMinute;
unsigned char timer10ms ;
unsigned char timer1sec;
unsigned int minutes;
char do_btu_flag;
char status;
char lan_watchdog;

void set_tx_buffer_crc();
uint16_t get_adc(uint8_t chan);
void calc_crc(char data);
void make_info_pkt();
void delay_100us(uint8_t count);

//--------------------------------------------------------------------------
/* Read the 10 bit analog to digital converter */

uint16_t get_adc(uint8_t chan)
{  
      
    outp(chan,ADMUX);               //Select channel
    sbi(ADCSR,ADSC);                //Start conversion 
    while( bit_is_clear(ADCSR,ADIF)); //Wait for ADC output registers to be loaded
    return __inw_atomic(ADCL);       //Return 16 bit value
}


//---------------------------------------------------------------------------

   // compute 16 bit CRC on 1 data byte. 
   //   crc is a  variable which accumulates the 16 bit CRC. 
   //   It must be cleared before starting to compute 16 bit CRC on a string.
   //   The final output value will be in crc.
   
void calc_crc(char data){

      uint8_t i;
     
      for (i=0;i<8;i++) {                //Do all 8 bits in the data byte
         if (data & 0x80) crc ^= 0x8000; //if data bit is 1 then flip crc msb
         data = data << 1;             // get next data bit
         if (crc & 0x8000) { //crc msb is 1
             crc = crc << 1;
             crc = crc ^ 0x8005;
          } else crc = crc << 1;
        
      }

}   

//-----------------------------------------------------------------------
void set_tx_buffer_crc(){
      int j = tx_buf[COUNT + 0] + 2; //get data size
      int i;

      crc = 0;
      for (i=0;i<j;i++) calc_crc(tx_buf[i]); //Compute crc16 over buffer
      tx_buf[i++] = (uint8_t)(crc >> 8);  //tack the crc bytes on the end of the buffer
      tx_buf[i++] = (uint8_t)(crc & 0xff);
      
}

   //-----------------------------------------------------------------------

   /* Build a packet */
  
   void make_info_pkt(){
      uint8_t i;
      i = ADDRESS;
      tx_buf[i] = 1;             // address of T-Zero battle bot
      i = DATA;
      tx_buf[i++] = 0x01;        //data type = 1
      tx_buf[i++] = (uint8_t)ad0;     //1
      tx_buf[i++] = (uint8_t)ad1;     //2
      tx_buf[i++] = (uint8_t)ad2;     //3
      tx_buf[i++] = (uint8_t)ad3;     //4
      tx_buf[i++] = matrix0;          //5
      tx_buf[i++] = matrix1;          //6
            
      tx_buf[COUNT] = i - DATA;       //Set byte count field to data length

      set_tx_buffer_crc();  //compute and add CRC16 bits to end of pkt
      
      txstate = 2;          //Tell scheduler it's ready to transmit
      outp(0x28,UCR);       //enable UART TX and interrupt 
      outp(0xff,UDR);       // Send a FF to get things started 
   }
      
//---------------------------------------------------------------------------

 int main(void)
{
  outp(BV(CS01), TCCR0);      /* use CLK/8 source for counter */
  
  outp(0xff, DDRD);           /* port D all outputs */
  outp(0xff, PORTD);          /* turn all leds off  */
  outp(0x00,DDRA);            //8 bits analog inputs
  outp(0,PORTA);              //Clear port A
  outp(0x85,ADCSR);           //Enable A to D converter with /32 conv clk.
  outp(0xf0,DDRC);            //Low nibble inputs, high nibble outputs on port C
  outp(0x00,PORTC);           //Clear port C bits

  outp(0x0b,TCCR2);           //clear on compare, Prescale = 32 on T/C 2
  outp(192,OCR2);             //T/C 2 compare register
  //sbi(TIMSK,OCIE2);           //Enable timer 2 compare interrupts
  
  outp(47,UBRR);              /* UART baud rate set to 9600 */
  txstate = 0;
  sei();                      /* enable global interrupts */

    
  for(;;){                   //MAIN LOOP

     
        
    /* Get A-D converter values */
    
    ad0 = get_adc(0) >> 2;      //Convert 10 bit value to 8 bits by right shifts
    ad1 = get_adc(1) >> 2;
    ad2 = get_adc(2) >> 2;
    ad3 = get_adc(3) >> 2;
    ad4 = get_adc(4) >> 2;

    if((ad4 & 0xff) < 166){
         cbi(PORTD,LED_YELLOW);  //Yellow LED on if battery less than 6.5 volts
    }else{
       sbi(PORTD,LED_YELLOW);
    }

    
    /* Scan the diode switch matrix in the logitec wingman rumble pad */
    
    outp(0x10,PORTC);            //Set matrix pin 5
    delay_100us(1);
    matrix0 = inp(PINC) & 0x0f;  //Get low 4 bits
   
    outp(0x20,PORTC);            //Set matrix pin 6
    delay_100us(1);
    val = inp(PINC) << 4;        //get low 4 bits and shift left 4
    matrix0 |= val;              //OR bits into matrix0

    outp(0x40,PORTC);            //Set matrix pin 7
    delay_100us(1);
    matrix1 = inp(PINC) & 0x0f;  //Get low 4 bits
    outp(0x80,PORTC);            //set matrix pin 8
    delay_100us(1);
    val = inp(PINC) << 4;        //get low 4 bits and shift left 4
    matrix1 |= val;              //or bits into matrix1
    outp(0x00,PORTC);            //Clear matrix pins
     
    
    while(txstate > 0) delay_100us(1) ;   //Wait for transmitter to be ready
    
    if(txstate == 0) make_info_pkt();     //Transmit data to robot


  }

}

//---------------------------------------


void delay_100us(uint8_t count)
{
  while(count){
    outp(256 - ((F_CPU/80000) -1), TCNT0);
    outp(BV(TOV0), TIFR);
    while(!(inp(TIFR) & BV(TOV0)));   
    count--;
  }
}


void delay_10ms(uint8_t count)
{
  while(count){
    delay_100us(100);
    count--;
  }
}

   

//-----------------------------------------------------------------------
/* Come here 1200 times per second */
SIGNAL(SIG_OUTPUT_COMPARE2)
{
   
   

      if(--prescale == 0){      //This executes 100 times per sec.
         prescale = 12;
         if(prescale1sec > 0) prescale1sec--;
         
      }

      
     if(prescale1sec == 0){  //This executes once per second
        prescale1sec = 100;
        if(prescaleMinute > 0) prescaleMinute--;

        
     }

     if(prescaleMinute == 0){ //This executes once per minute
        prescaleMinute = 60;
        minutes++;
        //lan_watchdog++;
        if(minutes >= 1440) minutes = 0; //Count for 24 hours and roll over
        //do_btu_flag = 1;
     }


}

//----------------------------------------------------------------------

/* UART interrupt handler */  
SIGNAL(SIG_UART_DATA)
{      
      switch(txstate) {
      case 0:  break;
      case 1:
      case 2:
      case 3:  outp(0xff,UDR) ; txstate++; cbi(PORTD,LED_RED);
         break;
      case 4:  outp(0xf0,UDR); txstate++;
         break;
      case 5:  outp(0x55,UDR); txstate++;
         break;

      case 6:  tx_ptr = 0;
         tx_ctr = tx_buf[COUNT] + 4; 
         txstate++;
         //fall through...

      case 7:  outp(tx_buf[tx_ptr++],UDR);  //Now send buffer contents
            tx_ctr--;
            if (tx_ctr == 0){ 
               txstate = 8;
               
            }
         
            break;

      case 8:   
               txstate = 0;
               outp(0xff,UDR);   //Feed the uart data register one last time to clear this interrupt
               outp(UCR,0);      //Stop the UART transmitter and Interrupts
               sbi(PORTD,LED_RED);     //LED off
               break;



      }

      
}


//---------------------------------------------------------------------------

/* end of file */



